import os
import requests


def get_pr(owner, repo, number):
    """Get pull request information"""
    return requests.get(
        f"https://api.github.com/repos/{owner}/{repo}/pulls/{number}",
        headers={
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28",
        },
    ).json()


def get_repo_and_branch_for_pr(owner, repo, number):
    """Return the fork repo and branch branch for a given pull request number"""
    response = get_pr(owner, repo, number)
    full_name = response["head"]["repo"]["full_name"]
    ref = response["head"]["ref"]
    url = f"https://github.com/{full_name}"
    return url, ref


def get_repo_and_branch_for_readthedocs(repository_url, default_branch="master"):
    """Get repository and branch for the current readthedocs build"""
    # TODO: handle the case where READTHEDOCS_VERSION_NAME == "stable"
    version_name = os.environ.get("READTHEDOCS_VERSION_NAME", default_branch)

    if version_name == "latest":
        version_name = default_branch

    print(f"Version: {version_name}")

    username, reponame = repository_url.split("/")[-2:]

    try:
        # check if a number (this means the build is from a PR)
        int(version_name)
    except ValueError:
        repository_branch = version_name
    else:
        repository_url, repository_branch = get_repo_and_branch_for_pr(
            username, reponame, version_name
        )

    print(f"URL: {repository_url}. Branch: {repository_branch}")

    return repository_url, repository_branch



"""
Managing project versions
"""
import sys
import shutil
import subprocess
from pathlib import Path
import click

from pkgmt.changelog import expand_github_from_changelog, CHANGELOG
from pkgmt.versioner.versionernonsetup import VersionerNonSetup
from pkgmt.versioner.versionersetup import VersionerSetup
from pkgmt.versioner.util import complete_version_string, is_pre_release
from pkgmt.deprecation import Deprecations


def replace_in_file(path_to_file, original, replacement):
    """Replace string in file"""
    with open(path_to_file, "r+") as f:
        content = f.read()
        updated = content.replace(original, replacement)
        f.seek(0)
        f.write(updated)
        f.truncate()


def call(*args, **kwargs):
    return subprocess.run(*args, **kwargs, check=True)


def delete_dirs(*dirs):
    for dir_ in dirs:
        if Path(dir_).exists():
            shutil.rmtree(dir_)


def _input(prompt):
    return input(prompt)


def input_str(prompt, default):
    separator = " " if len(prompt.splitlines()) == 1 else "\n"
    response = _input(prompt + f"{separator}(Default: {default}): ")

    if not response:
        response = default

    return response


def input_confirm(prompt, abort):
    separator = " " if len(prompt.splitlines()) == 1 else "\n"
    response_raw = _input(prompt + f"{separator}Confirm? [y/n]: ")
    response = response_raw in {"y", "Y", "yes"}

    if not response and abort:
        print("Aborted!")
        sys.exit(1)

    return response


def validate_version_string(version):
    if not len(version):
        raise ValueError(f"Got invalid empty version string: {version!r}")

    if version[0] not in "0123456789":
        raise ValueError(
            f"Got invalid version string: {version!r} "
            "(first character must be numeric)"
        )

    return complete_version_string(version)


def version(
    project_root=".", tag=True, version_package=None, yes=False, push=True, target=None
):
    """

    Parameters
    ----------
    tag : bool, default=True
        Tags the commit with the stable version

    yes : bool, default=False
        Skips user prompt before applying changes

    push : bool, default=True
        Pushes the changes to the remote repository

    target : {None, "stable"}, default=None
        If None, it assumes the repository is in a dev version, so it creates a
        stable version and a new dev version. If stable, it assumes the repo is in a
        dev version and creates a stable version (skipping bumping to a new dev
        version)


    Notes
    -----
    Create a new version (projects with setup.py) :
    1. version_package will be None
    2. Set new stable version in package_name/__init__.py
    3. Update header in CHANGELOG file, and ask to review CHANGELOG
    4. Create commit for new version, create git tag, and push
    5. Set new development version in package_name/__init__.py, and CHANGELOG
    6. Commit new development version and push

    Create a new version (projects without setup.py) :
    1. These projects should contain two essential files:
       config.yaml in root directory which should contain the repo
       name
       _version.py file containing __version__ in the required directory
    2. version_package will be the directory containing _version.py file
    3. Set new stable version in package_name/_version.py
    4. Update header in CHANGELOG file, and ask to review CHANGELOG
    5. Create commit for new version, create git tag, and push
    6. Set new development version in package_name/_version.py, and CHANGELOG
    7. Commit new development version and push
    """

    if version_package:
        versioner = VersionerNonSetup(version_package, project_root=project_root)
    else:
        versioner = VersionerSetup(project_root=project_root)

    changelog_md_exists = (
        versioner.path_to_changelog and versioner.path_to_changelog.suffix == ".md"
    )

    # TODO: make it compatible with VersionerNonSetup
    if not version_package and changelog_md_exists:
        # check changelog
        CHANGELOG.from_path(
            path=versioner.path_to_changelog, project_root=project_root
        ).check()

        # look for deprecations
        Deprecations(root_dir=project_root).check()

    current = versioner.current_version()
    release = versioner.release_version()

    if yes:
        print(f"Releasing version: {release}")
    else:
        release = input_str(
            "Current version in setup.py is {current}. Enter"
            " release version".format(current=current),
            default=release,
        )

    release = validate_version_string(release)

    if versioner.path_to_changelog and not is_pre_release(release):
        versioner.update_changelog_release(release)

        changelog = versioner.path_to_changelog.read_text()

        if not yes:
            input_confirm(
                f"\n{versioner.path_to_changelog} content:" f"\n\n{changelog}\n",
                abort=True,
            )

    # Expand github links and sort secions
    if changelog_md_exists:
        expand_github_from_changelog(path=versioner.path_to_changelog)

        # TODO: make it compatible with VersionerNonSetup
        if not version_package:
            # sort changelog entries
            changelod_sorted = CHANGELOG.from_path(
                path=versioner.path_to_changelog, project_root=project_root
            ).sort_last_section()
            Path(versioner.path_to_changelog).write_text(changelod_sorted)

    else:
        print("Skipping CHANGELOG processing (only supported in .md files)")

    # Replace version number and create tag
    print("Commiting release version: {}".format(release))
    versioner.commit_version(
        release,
        msg_template="{package_name} release {new_version}",
        tag=tag,
        push=push,
    )

    if target == "stable":
        print(f"Version {release} was created.")
        return

    # Create a new dev version and save it
    bumped_version = versioner.bump_up_version()

    if not is_pre_release(release):
        print("Creating new section in CHANGELOG...")
        versioner.add_changelog_new_dev_section(bumped_version)

    print("Commiting dev version: {}".format(bumped_version))
    versioner.commit_version(
        bumped_version,
        msg_template="Bumps up {package_name} to version {new_version}",
        tag=False,
    )

    if push:
        call(["git", "push", "--no-verify"])

    print("Version {} was created, you are now in {}".format(release, bumped_version))


def upload(tag, production):
    """
    Check outs a tag, uploads to PyPI
    """
    print("Checking out tag {}".format(tag))
    call(["git", "checkout", tag])

    versioner = VersionerSetup()
    current = versioner.current_version()

    input_confirm(
        "Version in {} tag is {}. Do you want to continue?".format(tag, current),
        abort=True,
    )

    # create distribution
    delete_dirs("dist", "build", f"{versioner.PACKAGE}.egg-info")
    call(["python", "setup.py", "bdist_wheel", "sdist"])

    print("Publishing to PyPI...")

    if not production:
        call(
            [
                "twine",
                "upload",
                "--repository-url",
                "https://test.pypi.org/legacy/",
                "dist/*",
            ]
        )
    else:
        call(["twine", "upload", "dist/*"])

    try:
        call(["git", "checkout", "main"])
    except subprocess.CalledProcessError:
        try_master = True
    else:
        try_master = False

    if try_master:
        try:
            call(["git", "checkout", "master"])
        except subprocess.CalledProcessError:
            pass

    if production:
        click.secho("Published to PyPI.", fg="green")
    else:
        click.secho("Published to PyPI test server.", fg="yellow")



import re
import abc
import ast
import datetime
import subprocess

from pathlib import Path

from pkgmt.versioner.util import complete_version_string


def replace_in_file(path_to_file, original, replacement):
    """Replace string in file"""
    with open(path_to_file, "r+") as f:
        content = f.read()
        updated = content.replace(original, replacement)
        f.seek(0)
        f.write(updated)
        f.truncate()


def call(*args, **kwargs):
    return subprocess.run(*args, **kwargs, check=True)


def make_header(content, path, add_date=False):
    if add_date:
        today = datetime.datetime.now().strftime("%Y-%m-%d")
        content += f" ({today})"

    if path.suffix == ".md":
        return f"## {content}"
    elif path.suffix == ".rst":
        return f"{content}\n" + "-" * len(content)
    else:
        raise ValueError("Unsupported format, must be .rst or .md")


class AbstractVersioner(abc.ABC):
    def __init__(self, project_root="."):
        self.project_root = project_root or "."
        self.package_name, self.PACKAGE = self.find_package()
        if Path(self.project_root, "CHANGELOG.rst").exists():
            self.path_to_changelog = Path(self.project_root, "CHANGELOG.rst")
        elif Path(self.project_root, "CHANGELOG.md").exists():
            self.path_to_changelog = Path(self.project_root, "CHANGELOG.md")
        else:
            self.path_to_changelog = None
        self.version_file = self.version_file()

    def current_version(self):
        """Returns the current version in version file"""
        _version_re = re.compile(r"__version__\s+=\s+(.*)")

        with open(self.PACKAGE / self.version_file, "rb") as f:
            VERSION = str(
                ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
            )

        return VERSION

    def release_version(self):
        """
        Returns a release version number
        e.g. 2.4.4dev -> v.2.2.4
        """
        current = self.current_version()

        if "dev" not in current:
            raise ValueError("Current version is not a dev version")

        new = current.replace("dev", "")

        return complete_version_string(new)

    def bump_up_version(self):
        """
        Gets a the current released version and returns the next value.
        e.g. 1.2.5 -> 1.2.6dev

        Notes
        -----
        If a doing a pre-release (e.g., 1.0b1), the new version returns to dev
        (e.g., 1.0b1 -> 1.0dev)
        """
        # Get current version
        current = self.current_version()

        # pre-releases
        for pre in ["a", "b", "rc"]:
            if pre in current:
                prefix = current.split(pre)[0]
                return f"{prefix}dev"

        if "dev" in current:
            raise ValueError(
                "Current version is dev version, new dev "
                "versions can only be made from release versions"
            )

        # Get Z from X.Y.Z and sum 1
        tokens = current.split(".")

        # if just released a major version, add a 0 so we bump up a subversion
        # e.g. from 0.8 -> 0.8.0, then new dev version becomes 0.8.1dev
        if len(tokens) == 2:
            tokens.append("0")

        new_subversion = int(tokens[-1]) + 1

        # Replace new_subversion in current version
        tokens[-1] = str(new_subversion)
        new_version = ".".join(tokens) + "dev"

        return new_version

    def commit_version(self, new_version, msg_template, tag=False, push=True):
        """
        Replaces version in  __init__ and optionally creates a tag in the git
        repository (also saves a commit)

        Parameters
        ----------
        tag: bool, default=False
            If True, it adds ``new_version`` as tag.

        push : bool, default=True
            If True, it pushes the commit with ``new_version`` and the tag. Ignored if
            ``tag=False``
        """
        current = self.current_version()

        # replace new version in version file
        replace_in_file(self.PACKAGE / self.version_file, current, new_version)

        # Run git add and git status
        print("Adding new changes to the repository...")
        call(["git", "add", "--all"])
        call(["git", "status"])

        # Commit repo with updated dev version
        print("Creating new commit release version...")
        msg = msg_template.format(
            package_name=self.package_name, new_version=new_version
        )
        call(["git", "commit", "-m", msg])

        # Create tag
        if tag:
            print("Creating tag {}...".format(new_version))
            message = msg_template.format(
                package_name=self.package_name, new_version=new_version
            )
            call(["git", "tag", "-a", new_version, "-m", message])

            if push:
                print("Pushing tags...")
                call(["git", "push", "origin", new_version, "--no-verify"])

    def update_changelog_release(self, new_version):
        """Updates changelog file, adding a new section"""
        current_version = self.current_version()

        # update CHANGELOG header
        header_current = make_header(
            current_version, self.path_to_changelog, add_date=False
        )

        header_new = make_header(new_version, self.path_to_changelog, add_date=True)

        replace_in_file(self.path_to_changelog, header_current, header_new)

    def add_changelog_new_dev_section(self, dev_version):
        if self.path_to_changelog:
            if self.path_to_changelog.suffix == ".rst":
                start_current = "CHANGELOG\n========="
            else:
                start_current = "# CHANGELOG"

            new_header = make_header(dev_version, self.path_to_changelog)
            start_new = f"{start_current}\n\n{new_header}"
            replace_in_file(self.path_to_changelog, start_current, start_new)
        else:
            print("No CHANGELOG.{rst,md} found, skipping changelog editing...")

    @abc.abstractmethod
    def find_package(self):
        pass

    @abc.abstractmethod
    def version_file(self):
        pass



import shutil
import tempfile
import os
import contextlib
from pathlib import Path

import click


def markdown(filename, inplace=True):
    # optional dependencies
    import jupytext
    import nbclient

    front_matter = """\
---
jupyter:
  jupytext:
    text_representation:
      extension: .md
      format_name: markdown
      format_version: '1.3'
      jupytext_version: 1.13.6
  kernelspec:
    display_name: Python 3 (ipykernel)
    language: python
    name: python3
---
"""

    # add front matter so jupytext correctly identifies the bash cells and adds
    # the %%bash magic
    content = front_matter + Path(filename).read_text()

    nb = jupytext.reads(content)

    click.echo(f"Running {filename}")

    if inplace:
        nbclient.execute(nb)
    else:
        dir_ = Path(tempfile.mkdtemp(), "files")
        click.echo(f"Creating tmp directory: {dir_}")

        click.echo("Copying files to tmp directory...")
        shutil.copytree(".", dir_)

        with chdir(dir_):
            nbclient.execute(nb)

    return nb


@contextlib.contextmanager
def chdir(directory):
    old_dir = os.getcwd()
    try:
        os.chdir(str(directory))
        yield
    finally:
        os.chdir(old_dir)



from unittest.mock import ANY
from pathlib import Path
import re
from enum import IntEnum
from functools import total_ordering
from copy import deepcopy
import click


try:
    import mistune
    from mistune.renderers.markdown import MarkdownRenderer
    from mistune.core import BlockState
    from mistune.inline_parser import InlineParser
except ModuleNotFoundError:
    mistune = None
    MarkdownRenderer = None
    BlockState = None
    InlineParser = object

from pkgmt import config
from pkgmt.versioner.versionersetup import VersionerSetup
from pkgmt.versioner import util
from pkgmt._format import pretty_iterator
from pkgmt.exceptions import ProjectValidationError

_PREFIXES = {"[API Change]", "[Feature]", "[Fix]", "[Doc]"}


class CustomInlineParser(InlineParser):
    """
    Custom parser to prevent escaping characters inside `code spans`
    """

    def parse_codespan(self, m, state) -> int:
        marker = m.group(0)
        # require same marker with same length at end

        pattern = re.compile(r"(.*?(?:[^`]))" + marker + r"(?!`)", re.S)

        pos = m.end()
        m = pattern.match(state.src, pos)
        if m:
            end_pos = m.end()
            code = m.group(1)
            # Line endings are treated like spaces
            code = code.replace("\n", " ")
            if len(code.strip()):
                if code.startswith(" ") and code.endswith(" "):
                    code = code[1:-1]
            # removed the escaping here
            state.append_token({"type": "codespan", "raw": code})
            return end_pos
        else:
            state.append_token({"type": "text", "raw": marker})
            return pos


def _replace_issue_number_with_links(url, text):
    # taken from jupytext/tests/test_changelog.py
    return re.sub(
        r"([^\[])#([0-9]+)",
        r"\1[#\2](" + url + r"\2)",
        text,
    )


def _expand_github_from_text(text):
    """Convert strings with the #{number} format into their"""
    cfg = config.load()
    url = f'https://github.com/{cfg["github"]}/issues/'
    return _replace_issue_number_with_links(url, text)


def expand_github_from_changelog(path="CHANGELOG.md"):
    path = Path(path)
    changelog = path.read_text()
    changelog_ = _expand_github_from_text(changelog)
    path.write_text(changelog_)


# functions for checking CHANGELOG contents


def _find_first_list_after(tree, idx):
    for idx, element in enumerate(tree[idx + 1 :], start=idx + 1):
        if element["type"] == "heading":
            return None, idx
        elif element["type"] == "list":
            return element, idx

    raise ProjectValidationError(
        "Error parsing CHANGELOG: Could not find a list after the H2 heading"
    )


def _extract_text_from_token(token):
    if token["type"] in {"text", "codespan"}:
        return token["raw"]
    elif token["type"] == "link":
        return token["children"][0]["raw"]
    else:
        raise NotImplementedError(f"Unable to extract text from token: {token}")


def _extract_text_from_item(item):
    return "".join(
        [_extract_text_from_token(ch) for ch in item["children"][0]["children"]]
    )


def _extract_text_from_items(list_):
    return [_extract_text_from_item(item) for item in list_["children"]]


def _valid_item(item):
    return any(item.startswith(prefix) for prefix in _PREFIXES)


def _breaks_api(items):
    return any(item.startswith("[API Change]") for item in items)


@total_ordering
class ListItem:
    def __init__(self, token) -> None:
        self.token = token
        self.text = "".join(
            [_extract_text_from_token(ch) for ch in token["children"][0]["children"]]
        )
        result = re.search(r"(\[API Change\]|\[Feature\]|\[Fix\]|\[Doc\])", self.text)
        self.type = result.group()
        self.type_enum = ListItemType.from_str(self.type)

    def __eq__(self, other):
        return self.type_enum == other.type_enum

    def __lt__(self, other):
        return self.type_enum < other.type_enum

    def __repr__(self) -> str:
        return f"LisItem({self.text!r})"


class ListItemType(IntEnum):
    api_change = 1
    feature = 2
    fix = 3
    doc = 4

    @classmethod
    def from_str(cls, s):
        mapping = {
            "[API Change]": 1,
            "[Feature]": 2,
            "[Fix]": 3,
            "[Doc]": 4,
        }
        return mapping[s]


class CHANGELOG:
    """Run several checks in the CHANGELOG.md file"""

    def __init__(self, text, project_root=".") -> None:
        if not mistune:
            raise ModuleNotFoundError(
                "Checking CHANGELOG.md requires mistune 3. "
                "Install it with: pip install 'pkgmt[check]'"
            )

        if mistune.__version__[0] != "3":
            raise ModuleNotFoundError(
                "Checking CHANGELOG.md requires mistune 3. "
                f"You have {mistune.__version__}. Install it with: "
                "pip install 'pkgmt[check]'"
            )

        self.text = text

        markdown = mistune.Markdown(
            renderer=None, inline=CustomInlineParser(hard_wrap=False), plugins=None
        )
        self.tree = markdown(text)

        versioner = VersionerSetup(project_root=project_root)
        self.current = versioner.current_version()

    @classmethod
    def from_path(cls, path, project_root="."):
        return cls(text=Path(path).read_text(), project_root=project_root)

    def sort_last_section(self):
        """Sorts last section depending on the prefix"""
        self.check_latest_changelog_entries()

        idx_subheading, _ = self.get_first_subheading()
        list_, idx = _find_first_list_after(self.tree, idx_subheading)
        renderer = MarkdownRenderer()

        if list_:
            li = sorted([ListItem(ch) for ch in list_["children"]])
            tree = deepcopy(self.tree)
            tree[idx]["children"] = [item.token for item in li]
            return renderer(tree, state=BlockState())
        else:
            # list is empty, nothing to sort
            return self.text

    def get_first_subheading(self):
        """Find the first H2 heading. Returns index (in the tree) and text"""
        subheading = {
            "type": "heading",
            "attrs": {"level": 2},
            "style": "axt",
            "children": [{"type": "text", "raw": ANY}],
        }

        for idx, element in enumerate(self.tree):
            if element == subheading:
                return idx, element["children"][0]["raw"]

        raise ProjectValidationError(
            "Error parsing CHANGELOG: Could not find an H2 heading"
        )

    def get_latest_changelog_section(self):
        """Gets the elements in the first section (below the first H2 heading)"""
        idx_subheading, _ = self.get_first_subheading()
        changes, _ = _find_first_list_after(self.tree, idx_subheading)

        if changes:
            items_text = _extract_text_from_items(changes)
            return items_text
        else:
            return []

    def check_consistent_dev_version(self):
        """
        Ensures that there is a major development version set if there's at least one
        [API Change] entry in the latest CHANGELOG section
        """
        changelog_entries = self.get_latest_changelog_section()

        if _breaks_api(changelog_entries) and not util.is_major_version(self.current):
            raise ProjectValidationError(
                "[Unexpected version] Expected a major version since there is at least "
                f"one [API Change] CHANGELOG entry, got minor version: {self.current!r}"
            )

    def check_latest_changelog_entries(self):
        """
        Ensures the entries in the latest CHANGELOG section have the right prefixes
        """
        invalid = []

        for item in self.get_latest_changelog_section():
            if not _valid_item(item):
                invalid.append(item)

        if invalid:
            _PREFIXES_OUT = pretty_iterator(_PREFIXES)
            invalid_out = pretty_iterator(invalid)
            raise ProjectValidationError(
                f"[Invalid CHANGELOG] Changelog items must start with one "
                f"of: {_PREFIXES_OUT}. Fix the following entries: {invalid_out}"
            )

        return True

    def check_consistent_changelog_and_version(self):
        """
        Check the latest section in CHANGELOG.md matches the __version__ string
        """
        _, subheading = self.get_first_subheading()

        # when making stable releases, the format is: "{version} {date}"
        # in such cases, we only compare the "{version}" part
        date_match = re.match(r"([\w\.]+)( \([0-9]{4}-[0-9]{2}-[0-9]{2}\))", subheading)

        if (date_match and date_match.group(1) != self.current) or (
            date_match is None and subheading != self.current
        ):
            raise ProjectValidationError(
                "[Inconsistent version] Version in  top section in "
                f"CHANGELOG is {subheading!r}, "
                f"but version in __init__.py is {self.current!r}, "
                "fix them so they match"
            )

    def check(self, *, verbose=False):
        """Runs all checks"""
        errors = []

        for function, name in (
            (self.check_latest_changelog_entries, "CHANGELOG entries prefixes"),
            (
                self.check_consistent_changelog_and_version,
                "CHANGELOG and __init__.py consistency",
            ),
            (self.check_consistent_dev_version, "consistent dev version"),
        ):
            try:
                function()
            except Exception as e:
                errors.append(str(e))
                if verbose:
                    click.secho(f"Checking {name}... ERROR!", fg="red")
            else:
                if verbose:
                    click.secho(f"Checking {name}... OK!", fg="green")

        if errors:
            errors_ = "\n\n".join(f"- {e}" for e in errors)
            raise ProjectValidationError(
                f"Found the following errors in the project:\n{errors_}"
            )



import os
from pathlib import Path
import shutil
import importlib.resources
from string import Template

from pkgmt import assets


def render_inplace(path, **kwargs):
    template = Template(path.read_text())
    path.write_text(template.safe_substitute(**kwargs))


def package(name):
    """Create a new package"""
    # .path doesn't work with directories
    with importlib.resources.path(assets, "__init__.py") as p:
        # so we trick it
        path = p.parent / "template"

    shutil.copytree(path, name)

    root = Path(name)

    for file in ("README.md", "setup.py", "tasks.py", "MANIFEST.in"):
        render_inplace(root / file, name=name)

    os.rename(root / "src" / "name", root / "src" / name)



from invoke import task


@task
def setup(c, version=None):
    """
    Setup dev environment, requires conda
    """
    version = version or "3.9"
    suffix = "" if version == "3.9" else version.replace(".", "")
    env_name = f"$name{suffix}"

    c.run(f"conda create --name {env_name} python={version} --yes")
    c.run(
        'eval "$(conda shell.bash hook)" '
        f"&& conda activate {env_name} "
        "&& pip install --editable .[dev]"
    )

    print(f"Done! Activate your environment with:\nconda activate {env_name}")


@task(aliases=["v"])
def version(c):
    """Create a new stable version commit"""
    from pkgmt import versioneer

    versioneer.version(project_root=".", tag=True)


@task(aliases=["r"])
def release(c, tag, production=True):
    """Upload to PyPI"""
    from pkgmt import versioneer

    versioneer.upload(tag, production=production)



import toml
import yaml
from pathlib import Path


def load():
    if Path("pyproject.toml").exists():
        with open("pyproject.toml") as f:
            return toml.load(f)["tool"]["pkgmt"]
    elif Path("config.yaml").exists():
        with open("config.yaml", "r") as stream:
            try:
                return yaml.safe_load(stream)["pkgmt"]
            except yaml.YAMLError as exc:
                print("Error loading config.yaml file : {}".format(exc))



"""
Check which dependencies were updated most recently
"""

import concurrent.futures
from functools import total_ordering

import click
import requests
from pip._internal.operations import freeze


def get_latest_version(package_name):
    """Check the latest available version"""
    res = requests.get(f"https://pypi.org/pypi/{package_name}/json")
    return res.json()


@total_ordering
class Package:
    def __init__(self, name) -> None:
        self.name = name
        self.last_version = None
        self.last_updated = None

    def fetch(self):
        response = get_latest_version(self.name)

        self.last_version = response["info"]["version"]

        # this returns one for wheel, one for source, and possibly
        # others for yanked versions (?), we use the most recent one
        self.last_updated = max(
            [r["upload_time_iso_8601"] for r in response["releases"][self.last_version]]
        )

    def __repr__(self) -> str:
        repr_ = f"{type(self).__name__}({self.name!r})"

        if self.last_updated:
            repr_ = f"{repr_} @ {self.last_updated}"

        return repr_

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.last_updated < other.last_updated


def main():
    pkgs = [Package(dep.split("==")[0]) for dep in freeze.freeze()]

    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        future2pkg = {executor.submit(pkg.fetch): pkg for pkg in pkgs}

        for future in concurrent.futures.as_completed(future2pkg):
            try:
                future.result()
            except Exception as e:
                pkg = future2pkg[future]
                print(f"{pkg} generated an exception: {e}")

    pkgs_valid = sorted([pkg for pkg in pkgs if pkg.last_updated is not None])

    pkgs_valid_ = "\n".join(repr(pkg) for pkg in pkgs_valid)

    pkgs_invalid = "\n".join([repr(pkg) for pkg in pkgs if pkg.last_updated is None])

    click.echo(f"Invalid packages: {pkgs_invalid}")
    click.echo(f"Valid packages (recently updated first): {pkgs_valid_}")


@click.command()
def _cli():
    """Retrieve latest updated packages in the current environment"""
    main()


if __name__ == "__main__":
    _cli()



import subprocess
from pathlib import Path
import stat
import click
import shutil
from contextlib import contextmanager


pre_push_hook = """\
#!/usr/bin/env python3

from pathlib import Path
import sys
import subprocess

try:
    import jupytext
except ModuleNotFoundError:
    jupytext = None


try:
    import nbqa
except ModuleNotFoundError:
    nbqa = None


def find_root():
    current = Path().resolve()

    while not (current / 'pyproject.toml').exists():
        if current == current.parent:
            sys.exit('Could not find project root.')

        current = current.parent

    return str(current)


current = find_root()

print("*** Running: flake8 ***")
res = subprocess.run(['flake8'], cwd=current)

error = False

if res.returncode:
    error = True

if not nbqa:
    print("nbqa is missing, flake8 won't run on notebooks. "
          "Fix it with: pip install nbqa")

if not jupytext:
    print("jupytext is missing, flake8 won't run on notebooks. "
          "Fix it with: pip install jupytext")


if nbqa and jupytext:
    print("*** Running: nbqa flake8 ***")
    res_nb = subprocess.run(["nbqa", "flake8", "."], cwd=current)

    if res_nb.returncode:
        error = True

if error:
    print()
    sys.exit('***flake8 returned errors. Fix and push again.***')

print("flake8 passed!")
"""


def _delete_hook(path):
    """Delete a git hook at the given path"""
    if path.exists():
        if path.is_file():
            path.unlink()
        else:
            # in the remote case that it's a directory
            shutil.rmtree(path)

    click.echo(f"Deleted hook located at {path}")


def _install_hook(path_to_hook, content):
    """
    Install a git hook script at the given path
    """
    if path_to_hook.exists():
        click.echo("Overwriting existing hook...")

    path_to_hook.write_text(content)
    # make the file executable
    path_to_hook.chmod(path_to_hook.stat().st_mode | stat.S_IEXEC)


def install_hook():
    if not Path(".git").is_dir():
        raise NotADirectoryError(
            "Expected a .git/ directory in the current working "
            "directory. Run this from the repository root directory."
        )

    parent = Path(".git", "hooks")
    parent.mkdir(exist_ok=True)

    _install_hook(parent / "pre-push", pre_push_hook)
    click.echo("Successfully installed pre-push git hook")


def uninstall_hook():
    _delete_hook(Path(".git", "hooks", "pre-push"))


def run_hook():
    """Run hook without installing it"""
    with tmp_file() as path:
        path.write_text(pre_push_hook)
        subprocess.run(["python", "_pkgmt_hook.py"])


@contextmanager
def tmp_file():
    path = Path("_pkgmt_hook.py")

    try:
        yield path
    finally:
        if path.exists():
            path.unlink()
